`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: University of Hawaii-Manoa
// Engineer: Patrick Allison
// 
// Create Date:    14:25:52 06/12/2008 
// Design Name: 
// Module Name:    GRAY_COUNTER 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: Fully pipelined 32-bit and 4-bit Gray code counter.
// The 4-bit was originally for testing, but it's relatively easy to change the
// bit length to any arbitrary number. The magic numbers would have to change, however.
// If I get ambitious at some point, I may make this fully parametrizable in terms
// of number of bits and include all of the math to generate the magic numbers
// automatically.
//
// This design is essentially infinitely extendable in number of bits,
// at the cost of resources or pipeline delay only. Each stage only connects
// to one other stage, so P&R should have no problem actually meeting the theoretical
// max given by synthesis. The static timing reports for this design have the
// constraining path being given 100% by logic (or the RESET line).
//
// The design can start at any arbitrary number when cleared, but the magic numbers
// are set up to start at zero currently. 
//
// A 32-bit Gray code counter takes up ~169 slices in a Xilinx chip. Most of this is
// due to the requirement that the counter puts out Gray code in a single clock cycle.
// If this was not necessary, only the registers for the Gray cell pipes would be needed,
// but then the output would have to latch bits over the next 32 clock cycles - i.e.
// clock 0 = latch bit 0, clock 1 = latch bit 1 .. clock 31 = latch bit 31, resulting
// in a pipeline delay of 32 clock cycles. In most cases this would easily be acceptable.
//
// A module implementing that would simply have to connect the required number of GCPIPEs
// together and initialize it properly.
//
// Also note that a free-running counter (if RESET is not generated) uses significantly less
// resources on a Xilinx chip because the pipeline stages can be converted into SRL16s,
// dividing the number of required slices in the pipeline stage by 32.
//
// Finally, note that when running at ridiculously high speeds, the propagation time across
// the chip is quite large. So a RESET pulse might take ~5 ns to propagate through,
// and so the "zero" point might be a few counts off. The RESET is synchronous because
// it makes no sense to have an asynchronous reset of a counter.
//
// The Gray cell (GCPIPE) design was taken from
// http://asicdigitaldesign.wordpress.com 's "Counting in Gray" articles.
//
// Dependencies: 
//
// Revision: 1.0 
// Revision 0.01 - File Created
// Revision 1.0  - Tested behaviorally & with actual delays. Runs at basically
//                 the maximum speed of the chip.
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////

// 16-bit Gray code counter Positive Edge
module GC12_POS(input PCLK,
				input RESET,
				output [11:0] GCODE
				);
	reg [11:0] QS00;
	reg [11:0] QS01;
	reg [11:0] QS02;
	reg [11:0] QS03;
	reg [11:0] QS04;
	reg [11:0] QS05;
	reg [11:0] QS06;
	reg [11:0] QS07;
	reg [11:0] QS08;
	reg [11:0] QS09;
	reg [11:0] QS10;
	wire [11:0] QS11;

	// Parity. This is the inverted parity (i.e. inverse of XOR of all bits)
	// of the starting number. So with a start of 0
	reg P;
	// Magic number-land.
	// These are the 16-bit Gray codes counting up to zero and through 15,
	// pipelined 16 times, so if we start at 0,
	//                bit 15 is the 15s bits of Gray(FFE1 -> 0),
	//                bit 14 is the 15s bits of Gray(FFE2 -> 1),
	//                ..
	//                bit  1 is the 1s bits of Gray(FFFF -> 14),
	//                bit  0 is the 0s bits of Gray(0 -> 15)	
	
	// Note that they're identical to the 16-bit versions, just with the
	// high bit set (before, bit 31 was set, now, bit 15 is set) and only
	// from 0 to 15.
	
	parameter INIT00=12'h800;
	parameter INIT01=12'h801;
	parameter INIT02=12'h801;
	parameter INIT03=12'h802;
	parameter INIT04=12'h802;
	parameter INIT05=12'h803;
	parameter INIT06=12'h807;
	parameter INIT07=12'h804;
	parameter INIT08=12'h804;
	parameter INIT09=12'h805;
	parameter INIT10=12'h805;
	parameter INIT11=12'h00E;

	
	// ZINIT is a super-magic number.
	// "ffs()" is a libc function which returns the first set bit, starting from
	// 1 (so ffs(1) = 1, ffs(8) = 4, etc.).
	// Z[0] and Z[1] are actually unused and unimportant.
	// Z[2] is ffs(Gray(start+29)) >= 3
	// Z[3] is ffs(Gray(start+28)) >= 4
	// Z[4] is ffs(Gray(start+27)) >= 5
	// etc.
	
	// Luckily, for starting at zero, it's easy. They're all zero.
	parameter ZINIT11=32'h00000000;
	
	initial begin
		QS00 <= INIT00; 
		QS01 <= INIT01; 
		QS02 <= INIT02; 
		QS03 <= INIT03; 

		QS04 <= INIT04; 
		QS05 <= INIT05; 
		QS06 <= INIT06; 
		QS07 <= INIT07; 

		QS08 <= INIT08;
		QS09 <= INIT09;
		QS10 <= INIT10;
				
		P <= 1;
	end
	
	defparam gcstage.QINIT=INIT11,gcstage.ZINIT=ZINIT11;
	GCPIPESTAGE_12_POS gcstage(.PCLK(PCLK),.RESET(RESET),.QO(QS11), .P(P));
	
	always @(posedge PCLK) begin
		if (RESET) begin
			QS00 <= INIT00; 
			QS01 <= INIT01; 
			QS02 <= INIT02; 
			QS03 <= INIT03; 

			QS04 <= INIT04; 
			QS05 <= INIT05; 
			QS06 <= INIT06; 
			QS07 <= INIT07; 

			QS08 <= INIT08;
			QS09 <= INIT09;
			QS10 <= INIT10;
				
			P <= 1; // RIGHT
		end else begin
			QS10 <= QS11;
			QS09 <= QS10;
			QS08 <= QS09;
			QS07 <= QS08;

			QS06 <= QS07;
			QS05 <= QS06;
			QS04 <= QS05;
			QS03 <= QS04;

			QS02 <= QS03;
			QS01 <= QS02;
			QS00 <= QS01;
	
			P <= ~P;
		end
	end

	assign GCODE[11] = QS11[11];
	assign GCODE[10] = QS10[10];
	assign GCODE[09] = QS09[09];
	assign GCODE[08] = QS08[08];

	assign GCODE[07] = QS07[07];
	assign GCODE[06] = QS06[06];
	assign GCODE[05] = QS05[05];
	assign GCODE[04] = QS04[04];

	assign GCODE[03] = QS03[03];
	assign GCODE[02] = QS02[02];
	assign GCODE[01] = QS01[01];
	assign GCODE[00] = QS00[00];
endmodule



	
// 16-bit Gray code counter stage Positive Edge
module GCPIPESTAGE_12_POS(
	input PCLK,
	input RESET,
	output [11:0] QO,
	output [11:0] ZO,
	input P
	);
	parameter QINIT = 12'h000;
	parameter ZINIT = 12'h000;
		
	defparam pipe00.initQ = QINIT[00], pipe00.initZ = ZINIT[00];
	defparam pipe01.initQ = QINIT[01], pipe01.initZ = ZINIT[01];
	defparam pipe02.initQ = QINIT[02], pipe02.initZ = ZINIT[02];
	defparam pipe03.initQ = QINIT[03], pipe03.initZ = ZINIT[03];
	defparam pipe04.initQ = QINIT[04], pipe04.initZ = ZINIT[04];
	defparam pipe05.initQ = QINIT[05], pipe05.initZ = ZINIT[05];
	defparam pipe06.initQ = QINIT[06], pipe06.initZ = ZINIT[06];
	defparam pipe07.initQ = QINIT[07], pipe07.initZ = ZINIT[07];
	defparam pipe08.initQ = QINIT[08], pipe08.initZ = ZINIT[08];
	defparam pipe09.initQ = QINIT[09], pipe09.initZ = ZINIT[09];
	defparam pipe10.initQ = QINIT[10], pipe10.initZ = ZINIT[10];
	defparam pipe11.initQ = QINIT[11], pipe11.initZ = ZINIT[11];
		
	GCPIPE_12_POS pipe00(.PCLK(PCLK),.RESET(RESET),.Q_i(1'b1),.Q_o(QO[0]),.Z_i(1'b1), .P(P));
	GCPIPE_12_POS pipe01(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[0]),.Q_o(QO[1]),.Z_i(1'b1),.Z_o(ZO[1]),.P(P));
	GCPIPE_12_POS pipe02(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[1]),.Q_o(QO[2]),.Z_i(ZO[1]),.Z_o(ZO[2]),.P(~P));
	GCPIPE_12_POS pipe03(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[2]),.Q_o(QO[3]),.Z_i(ZO[2]),.Z_o(ZO[3]),.P(P));
	GCPIPE_12_POS pipe04(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[3]),.Q_o(QO[4]),.Z_i(ZO[3]),.Z_o(ZO[4]),.P(~P));
	GCPIPE_12_POS pipe05(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[4]),.Q_o(QO[5]),.Z_i(ZO[4]),.Z_o(ZO[5]),.P(P));
	GCPIPE_12_POS pipe06(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[5]),.Q_o(QO[6]),.Z_i(ZO[5]),.Z_o(ZO[6]),.P(~P));
	GCPIPE_12_POS pipe07(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[6]),.Q_o(QO[7]),.Z_i(ZO[6]),.Z_o(ZO[7]),.P(P));
	GCPIPE_12_POS pipe08(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[7]),.Q_o(QO[8]),.Z_i(ZO[7]),.Z_o(ZO[8]),.P(~P));
	GCPIPE_12_POS pipe09(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[8]),.Q_o(QO[9]),.Z_i(ZO[8]),.Z_o(ZO[9]),.P(P));
	GCPIPE_12_POS pipe10(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[9]),.Q_o(QO[10]),.Z_i(ZO[9]),.Z_o(ZO[10]),.P(~P));
	GCPIPE_12_POS pipe11(.PCLK(PCLK),.RESET(RESET),.Q_i(1'b1),.Q_o(QO[11]),.Z_i(ZO[10]),.P(P));

endmodule


module GCPIPE_12_POS(
	input PCLK,
	input RESET,
	input Q_i,
	output Q_o,
	input Z_i,
	output Z_o,
	input P
	);
	parameter initQ = 1'b0;
	parameter initZ = 1'b0;

	reg Q_val;
	reg Z_val;
	initial begin
		Z_val <= initZ;
		Q_val <= initQ;
	end
	
	always @(posedge PCLK) begin
		if (RESET) begin
			Z_val <= initZ;
		end else Z_val <= (!Q_i && Z_i);
	end
	assign Z_o = Z_val;
	
	always @(posedge PCLK) begin
		if (RESET) begin
			Q_val <= initQ;
		end else if (~P) begin
			Q_val <= Q_val ^ (Z_i && Q_i);
		end
	end
	assign Q_o = Q_val;
	
endmodule


// 16-bit Gray code counter Negative Edge
module GC12_NEG(input PCLK,
				input RESET,
				output [11:0] GCODE
				);
	reg [11:0] QS00;
	reg [11:0] QS01;
	reg [11:0] QS02;
	reg [11:0] QS03;
	reg [11:0] QS04;
	reg [11:0] QS05;
	reg [11:0] QS06;
	reg [11:0] QS07;
	reg [11:0] QS08;
	reg [11:0] QS09;
	reg [11:0] QS10;
	wire [11:0] QS11;

	// Parity. This is the inverted parity (i.e. inverse of XOR of all bits)
	// of the starting number. So with a start of 0
	reg P;
	// Magic number-land.
	// These are the 16-bit Gray codes counting up to zero and through 15,
	// pipelined 16 times, so if we start at 0,
	//                bit 15 is the 15s bits of Gray(FFE1 -> 0),
	//                bit 14 is the 15s bits of Gray(FFE2 -> 1),
	//                ..
	//                bit  1 is the 1s bits of Gray(FFFF -> 14),
	//                bit  0 is the 0s bits of Gray(0 -> 15)	
	
	// Note that they're identical to the 16-bit versions, just with the
	// high bit set (before, bit 31 was set, now, bit 15 is set) and only
	// from 0 to 15.
	
	parameter INIT00=12'h800;
	parameter INIT01=12'h801;
	parameter INIT02=12'h801;
	parameter INIT03=12'h802;
	parameter INIT04=12'h802;
	parameter INIT05=12'h803;
	parameter INIT06=12'h807;
	parameter INIT07=12'h804;
	parameter INIT08=12'h804;
	parameter INIT09=12'h805;
	parameter INIT10=12'h805;
	parameter INIT11=12'h00E;

	
	// ZINIT is a super-magic number.
	// "ffs()" is a libc function which returns the first set bit, starting from
	// 1 (so ffs(1) = 1, ffs(8) = 4, etc.).
	// Z[0] and Z[1] are actually unused and unimportant.
	// Z[2] is ffs(Gray(start+29)) >= 3
	// Z[3] is ffs(Gray(start+28)) >= 4
	// Z[4] is ffs(Gray(start+27)) >= 5
	// etc.
	
	// Luckily, for starting at zero, it's easy. They're all zero.
	parameter ZINIT11=32'h00000000;
	
	initial begin
		QS00 <= INIT00; 
		QS01 <= INIT01; 
		QS02 <= INIT02; 
		QS03 <= INIT03; 

		QS04 <= INIT04; 
		QS05 <= INIT05; 
		QS06 <= INIT06; 
		QS07 <= INIT07; 

		QS08 <= INIT08;
		QS09 <= INIT09;
		QS10 <= INIT10;
				
		P <= 1;
	end
	
	defparam gcstage.QINIT=INIT11,gcstage.ZINIT=ZINIT11;
	GCPIPESTAGE_12_NEG gcstage(.PCLK(PCLK),.RESET(RESET),.QO(QS11), .P(P));
	
	always @(negedge PCLK) begin
		if (RESET) begin
			QS00 <= INIT00; 
			QS01 <= INIT01; 
			QS02 <= INIT02; 
			QS03 <= INIT03; 

			QS04 <= INIT04; 
			QS05 <= INIT05; 
			QS06 <= INIT06; 
			QS07 <= INIT07; 

			QS08 <= INIT08;
			QS09 <= INIT09;
			QS10 <= INIT10;
				
			P <= 1; // RIGHT
		end else begin
			QS10 <= QS11;
			QS09 <= QS10;
			QS08 <= QS09;
			QS07 <= QS08;

			QS06 <= QS07;
			QS05 <= QS06;
			QS04 <= QS05;
			QS03 <= QS04;

			QS02 <= QS03;
			QS01 <= QS02;
			QS00 <= QS01;
	
			P <= ~P;
		end
	end

	assign GCODE[11] = QS11[11];
	assign GCODE[10] = QS10[10];
	assign GCODE[09] = QS09[09];
	assign GCODE[08] = QS08[08];

	assign GCODE[07] = QS07[07];
	assign GCODE[06] = QS06[06];
	assign GCODE[05] = QS05[05];
	assign GCODE[04] = QS04[04];

	assign GCODE[03] = QS03[03];
	assign GCODE[02] = QS02[02];
	assign GCODE[01] = QS01[01];
	assign GCODE[00] = QS00[00];
endmodule



	
// 16-bit Gray code counter stage Negative Edge
module GCPIPESTAGE_12_NEG(
	input PCLK,
	input RESET,
	output [11:0] QO,
	output [11:0] ZO,
	input P
	);
	parameter QINIT = 12'h000;
	parameter ZINIT = 12'h000;
		
	defparam pipe00.initQ = QINIT[00], pipe00.initZ = ZINIT[00];
	defparam pipe01.initQ = QINIT[01], pipe01.initZ = ZINIT[01];
	defparam pipe02.initQ = QINIT[02], pipe02.initZ = ZINIT[02];
	defparam pipe03.initQ = QINIT[03], pipe03.initZ = ZINIT[03];
	defparam pipe04.initQ = QINIT[04], pipe04.initZ = ZINIT[04];
	defparam pipe05.initQ = QINIT[05], pipe05.initZ = ZINIT[05];
	defparam pipe06.initQ = QINIT[06], pipe06.initZ = ZINIT[06];
	defparam pipe07.initQ = QINIT[07], pipe07.initZ = ZINIT[07];
	defparam pipe08.initQ = QINIT[08], pipe08.initZ = ZINIT[08];
	defparam pipe09.initQ = QINIT[09], pipe09.initZ = ZINIT[09];
	defparam pipe10.initQ = QINIT[10], pipe10.initZ = ZINIT[10];
	defparam pipe11.initQ = QINIT[11], pipe11.initZ = ZINIT[11];
		
	GCPIPE_12_NEG pipe00(.PCLK(PCLK),.RESET(RESET),.Q_i(1'b1),.Q_o(QO[0]),.Z_i(1'b1), .P(P));
	GCPIPE_12_NEG pipe01(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[0]),.Q_o(QO[1]),.Z_i(1'b1),.Z_o(ZO[1]),.P(P));
	GCPIPE_12_NEG pipe02(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[1]),.Q_o(QO[2]),.Z_i(ZO[1]),.Z_o(ZO[2]),.P(~P));
	GCPIPE_12_NEG pipe03(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[2]),.Q_o(QO[3]),.Z_i(ZO[2]),.Z_o(ZO[3]),.P(P));
	GCPIPE_12_NEG pipe04(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[3]),.Q_o(QO[4]),.Z_i(ZO[3]),.Z_o(ZO[4]),.P(~P));
	GCPIPE_12_NEG pipe05(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[4]),.Q_o(QO[5]),.Z_i(ZO[4]),.Z_o(ZO[5]),.P(P));
	GCPIPE_12_NEG pipe06(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[5]),.Q_o(QO[6]),.Z_i(ZO[5]),.Z_o(ZO[6]),.P(~P));
	GCPIPE_12_NEG pipe07(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[6]),.Q_o(QO[7]),.Z_i(ZO[6]),.Z_o(ZO[7]),.P(P));
	GCPIPE_12_NEG pipe08(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[7]),.Q_o(QO[8]),.Z_i(ZO[7]),.Z_o(ZO[8]),.P(~P));
	GCPIPE_12_NEG pipe09(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[8]),.Q_o(QO[9]),.Z_i(ZO[8]),.Z_o(ZO[9]),.P(P));
	GCPIPE_12_NEG pipe10(.PCLK(PCLK),.RESET(RESET),.Q_i(QO[9]),.Q_o(QO[10]),.Z_i(ZO[9]),.Z_o(ZO[10]),.P(~P));
	GCPIPE_12_NEG pipe11(.PCLK(PCLK),.RESET(RESET),.Q_i(1'b1),.Q_o(QO[11]),.Z_i(ZO[10]),.P(P));

endmodule


module GCPIPE_12_NEG(
	input PCLK,
	input RESET,
	input Q_i,
	output Q_o,
	input Z_i,
	output Z_o,
	input P
	);
	parameter initQ = 1'b0;
	parameter initZ = 1'b0;

	reg Q_val;
	reg Z_val;
	initial begin
		Z_val <= initZ;
		Q_val <= initQ;
	end
	
	always @(negedge PCLK) begin
		if (RESET) begin
			Z_val <= initZ;
		end else Z_val <= (!Q_i && Z_i);
	end
	assign Z_o = Z_val;
	
	always @(negedge PCLK) begin
		if (RESET) begin
			Q_val <= initQ;
		end else if (~P) begin
			Q_val <= Q_val ^ (Z_i && Q_i);
		end
	end
	assign Q_o = Q_val;
	
endmodule
